//
// Created by dawsh on 1/17/24.
//

#include <JGL/JGL.h>
#include <glad/glad.h>
#include <J3ML/Algorithm/Bezier.hpp>
#include <JGL/logger/logger.h>

JGL::RenderTarget* render_target = nullptr;
GLfloat oldColor[4] = {0, 0, 0, 1};
GLfloat baseColor[4] = {1, 1, 1, 1};

GLuint current_fbo = 0;
GLint viewport[4] = {0, 0, 0, 0};

bool inJ2D = false;
bool inJ3D = false;
bool wasTexture2DEnabled = false;
bool wasTextureCoordArrayEnabled = false;
bool wasDepthTestEnabled = false;
bool wasVertexArraysEnabled = false;
bool wasCullFaceEnabled = false;
bool wasBlendEnabled = false;
bool wasColorArrayEnabled = false;
GLint activeTextureUnit = 0;

namespace JGL {
    using namespace J3ML;
    Vector2 wS;

    void Update(const Vector2& window_size) {
        wS = window_size;
        glViewport(0, 0, (int) wS.x, (int) wS.y);
    }

    bool MeetsRequirements() {
        if (!GLAD_GL_VERSION_2_1)
            return false;
        if (!GLAD_GL_EXT_framebuffer_object)
            return false;
        if (!GLAD_GL_ARB_framebuffer_object)
            return false;
        return true;
    }

#pragma region J2D
    void J2D::Begin(RenderTarget* rt, bool clear_buffers) {
        GLfloat old_clear_color[4];
        if (rt != nullptr) {
            render_target = rt;
            glGetFloatv(GL_COLOR_CLEAR_VALUE, old_clear_color);
            glGetIntegerv(GL_VIEWPORT, viewport);
            current_fbo = JGL::RenderTarget::GetActiveGLFramebufferHandle();
            JGL::RenderTarget::SetActiveGLRenderTarget(*rt);
        }
        glMatrixMode(GL_PROJECTION);
        glPushMatrix();
        glLoadIdentity();

        if (rt == nullptr)
            glViewport(0, 0, (int) wS.x, (int) wS.y),
            glOrtho(0, wS.x, wS.y, 0, -1, 1);
        else
            glOrtho(0, rt->GetDimensions().x, rt->GetDimensions().y, 0, -1, 1);

        glMatrixMode(GL_MODELVIEW);
        glPushMatrix();
        glLoadIdentity();

        //Get what the draw color was before we did anything.
        glGetFloatv(GL_CURRENT_COLOR, oldColor);
        glColor4fv(baseColor);

        glGetIntegerv(GL_ACTIVE_TEXTURE, &activeTextureUnit);
        activeTextureUnit = activeTextureUnit - GL_TEXTURE0;

        if (activeTextureUnit != 0)
            glActiveTexture(GL_TEXTURE0);

        if (glIsEnabled(GL_DEPTH_TEST))
            wasDepthTestEnabled = true,
            glDisable(GL_DEPTH_TEST);
        else
            wasDepthTestEnabled = false;

        if (!glIsEnabled(GL_VERTEX_ARRAY))
            wasVertexArraysEnabled = false,
            glEnableClientState(GL_VERTEX_ARRAY);
        else
            wasVertexArraysEnabled = true;

        if (!glIsEnabled(GL_CULL_FACE))
            wasCullFaceEnabled = false,
            glEnable(GL_CULL_FACE),
            glCullFace(GL_BACK);
        else
            wasCullFaceEnabled = true;

        if (!glIsEnabled(GL_BLEND))
            wasBlendEnabled = false,
            glEnable(GL_BLEND),
            glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
        else
            wasBlendEnabled = true;

        if (!glIsEnabled(GL_TEXTURE_2D))
            wasTexture2DEnabled = false,
            glEnable(GL_TEXTURE_2D);
        else
            wasTexture2DEnabled = true;

        if (!glIsEnabled(GL_TEXTURE_COORD_ARRAY))
            wasTextureCoordArrayEnabled = false,
            glEnableClientState(GL_TEXTURE_COORD_ARRAY);
        else
            wasTextureCoordArrayEnabled = true;

        if (glIsEnabled(GL_COLOR_ARRAY))
            wasColorArrayEnabled = true,
            glDisableClientState(GL_COLOR_ARRAY);
        else
            wasColorArrayEnabled = false;

        if (!inJ3D)
            inJ2D = true;
        else { Logger::Error("Attempt to Begin J2D inside of J3D context."); }

        if (rt != nullptr && clear_buffers) {
            glClearColor(rt->GetClearColor().R(), rt->GetClearColor().G(), rt->GetClearColor().B(), rt->GetClearColor().A());
            glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
            glClearColor(old_clear_color[0], old_clear_color[1], old_clear_color[2], old_clear_color[3]);
        }
    }

    void J2D::End() {
        //Change back to the previous projection (The 3D one in Re3D's case.)
        glPopMatrix();
        glMatrixMode(GL_PROJECTION);
        glPopMatrix();
        glMatrixMode(GL_MODELVIEW);
        if (wasDepthTestEnabled)
            glEnable(GL_DEPTH_TEST);

        if (!wasVertexArraysEnabled)
            glDisableClientState(GL_VERTEX_ARRAY);

        if (!wasCullFaceEnabled)
            glDisable(GL_CULL_FACE);

        if (!wasBlendEnabled)
            glDisable(GL_BLEND);

        if (!wasTexture2DEnabled)
            glDisable(GL_TEXTURE_2D);

        if (!wasTextureCoordArrayEnabled)
            glDisableClientState(GL_TEXTURE_COORD_ARRAY);

        if (wasColorArrayEnabled)
            glEnableClientState(GL_COLOR_ARRAY);

        //Select whatever texture_handle unit was selected before.
        glActiveTexture(GL_TEXTURE0 + activeTextureUnit);

        //Put the draw color back how it was before.
        glColor4fv(oldColor);

        if (render_target != nullptr) {
            render_target = nullptr;
            glBindFramebuffer(GL_FRAMEBUFFER, current_fbo);
            glViewport(viewport[0], viewport[1], viewport[2], viewport[3]);
        }
        inJ2D = false;
    }

    void J2D::DrawPoint(const Color4& color, const Vector2& coordinates, float radius) {
        if (!inJ2D)
            Logger::Error("Drawing J2D element before J2D begin.");

        glPointSize(radius);
        glColor4ubv(color.ptr());
        glVertexPointer(2, GL_FLOAT, sizeof(Vector2), coordinates.ptr());
        glDrawArrays(GL_POINTS, 0, 1);
        glColor4fv(baseColor);
    }

    void J2D::DrawPoint(const Color4& color, float x, float y, float radius) {
        DrawPoint(color, {x, y}, radius);
    }

    void J2D::DrawLine(const Color4& color, const Vector2& A, const Vector2& B, float thickness) {
        if (!inJ2D)
            Logger::Error("Drawing J2D element before J2D begin.");
        Vector2 vertices[] = {A, B};

        glLineWidth(thickness);
        glColor4ubv(color.ptr());
        glVertexPointer(2, GL_FLOAT, sizeof(Vector2), vertices);
        glDrawArrays(GL_LINES, 0, 2);
        glColor4fv(baseColor);
    }

    void J2D::DrawLine(const Color4& color, float x, float y, float w, float h, float thickness) {
        J2D::DrawLine(color, {x, y}, {w, h}, thickness);
    }

    void J2D::DrawGradientLine(const Color4& color1, const Color4& color2, const Vector2& A, const Vector2& B, float thickness) {
        if (!inJ2D)
            Logger::Error("Drawing J2D element before J2D begin.");

        Vector2 vertices[] = {A, B};
        GLfloat colors[8] = {color1.RedChannelNormalized(), color1.GreenChannelNormalized(), color1.BlueChannelNormalized(), color1.AlphaChannelNormalized(),
                             color2.RedChannelNormalized(), color2.GreenChannelNormalized(), color2.BlueChannelNormalized(), color2.AlphaChannelNormalized() };

        glEnableClientState(GL_COLOR_ARRAY);
        glLineWidth(thickness);
        glColorPointer(4,GL_FLOAT,sizeof(Color4), colors);
        glVertexPointer(2, GL_FLOAT, sizeof(Vector2), vertices);
        glDrawArrays(GL_LINES, 0, 2);
        glDisableClientState(GL_COLOR_ARRAY);
        glColor4fv(baseColor);
    }

    void DrawGradientLine(const Color4& color1, const Color4& color2, float x, float y, float w, float h, float thickness) {
        J2D::DrawGradientLine(color1, color2, {x, y}, {w, h}, thickness);
    }

    void J2D::OutlineRect(const Color4& color, const Vector2& pos, const Vector2& size, float thickness) {
        if (!inJ2D)
            Logger::Error("Drawing J2D element before J2D begin.");

        Vector2 vertices[] = {{pos.x, pos.y}, {pos.x, pos.y + size.y}, {pos.x + size.x, pos.y + size.y}, {pos.x + size.x, pos.y}};

        glLineWidth(thickness);
        glColor4ubv(color.ptr());
        glVertexPointer(2, GL_FLOAT, sizeof(Vector2), vertices);
        glDrawArrays(GL_LINE_LOOP, 0, 4);
        glColor4fv(baseColor);
    }

    void J2D::FillRect(const Color4& color, const Vector2& pos, const Vector2& size) {
        if (!inJ2D)
            Logger::Error("Drawing J2D element before J2D begin.");

        Vector2 vertices[] = {{pos.x, pos.y}, {pos.x, pos.y + size.y}, {pos.x + size.x, pos.y + size.y}, {pos.x + size.x, pos.y}};
        glColor4ubv(color.ptr());
        glVertexPointer(2, GL_FLOAT, sizeof(Vector2), vertices);
        glDrawArrays(GL_QUADS, 0, 4);
        glColor4fv(baseColor);
    }

    void J2D::FillGradientRect(const Color4& color1, const Color4& color2, const Direction& gradient, const Vector2& pos, const Vector2& size) {
        if (!inJ2D)
            Logger::Error("Drawing J2D element before J2D begin.");

        Vector2 vertices[] = {{pos.x, pos.y}, {pos.x, pos.y + size.y}, {pos.x + size.x, pos.y + size.y}, {pos.x + size.x, pos.y}};
        std::vector<GLfloat> colors{};

        if (gradient == Direction::Horizontal)
            colors = {color1.RedChannelNormalized(), color1.GreenChannelNormalized(), color1.BlueChannelNormalized(), color1.AlphaChannelNormalized(), color1.RedChannelNormalized(), color1.GreenChannelNormalized(), color1.BlueChannelNormalized(), color1.AlphaChannelNormalized(),
                      color2.RedChannelNormalized(), color2.GreenChannelNormalized(), color2.BlueChannelNormalized(), color2.AlphaChannelNormalized(), color2.RedChannelNormalized(), color2.GreenChannelNormalized(), color2.BlueChannelNormalized(), color2.AlphaChannelNormalized()};

        else if (gradient == Direction::Vertical)
            colors = {color1.RedChannelNormalized(), color1.GreenChannelNormalized(), color1.BlueChannelNormalized(), color1.AlphaChannelNormalized(), color2.RedChannelNormalized(), color2.GreenChannelNormalized(), color2.BlueChannelNormalized(), color2.AlphaChannelNormalized(),
                      color2.RedChannelNormalized(), color2.GreenChannelNormalized(), color2.BlueChannelNormalized(), color2.AlphaChannelNormalized(), color1.RedChannelNormalized(), color1.GreenChannelNormalized(), color1.BlueChannelNormalized(), color1.AlphaChannelNormalized()};

        else if (gradient == Direction::Diagonal_SWNE)
            colors = {(color1.RedChannelNormalized() + color2.RedChannelNormalized()) / 2.f, (color1.GreenChannelNormalized() + color2.GreenChannelNormalized()) / 2.f, (color1.BlueChannelNormalized() + color2.BlueChannelNormalized()) / 2.f, (color1.AlphaChannelNormalized() + color2.AlphaChannelNormalized()) / 2.f,
                      color1.RedChannelNormalized(), color1.GreenChannelNormalized(), color1.BlueChannelNormalized(), color1.AlphaChannelNormalized(), (color1.RedChannelNormalized() + color2.RedChannelNormalized()) / 2.f, (color1.GreenChannelNormalized() + color2.GreenChannelNormalized()) / 2.f,
                      (color1.BlueChannelNormalized() + color2.BlueChannelNormalized()) / 2.f, (color1.AlphaChannelNormalized() + color2.AlphaChannelNormalized()) / 2.f, color2.RedChannelNormalized(), color2.GreenChannelNormalized(), color2.BlueChannelNormalized(), color2.AlphaChannelNormalized()};

        else if (gradient == Direction::Diagonal_NWSE)
            colors = {color1.RedChannelNormalized(), color1.GreenChannelNormalized(), color1.BlueChannelNormalized(), color1.AlphaChannelNormalized(),(color1.RedChannelNormalized() + color2.RedChannelNormalized()) / 2.f, (color1.GreenChannelNormalized() + color2.GreenChannelNormalized()) / 2.f,
                      (color1.BlueChannelNormalized() + color2.BlueChannelNormalized()) / 2.f, (color1.AlphaChannelNormalized() + color2.AlphaChannelNormalized()) / 2.f, color2.RedChannelNormalized(), color2.GreenChannelNormalized(), color2.BlueChannelNormalized(), color2.AlphaChannelNormalized(),
                      (color1.RedChannelNormalized() + color2.RedChannelNormalized()) / 2.f, (color1.GreenChannelNormalized() + color2.GreenChannelNormalized()) / 2.f, (color1.BlueChannelNormalized() + color2.BlueChannelNormalized()) / 2.f,(color1.AlphaChannelNormalized() + color2.AlphaChannelNormalized()) / 2.f};

        glEnableClientState(GL_COLOR_ARRAY);
        glVertexPointer(2, GL_FLOAT, sizeof(Vector2), vertices);
        glColorPointer(4, GL_FLOAT, 0, colors.data());
        glDrawArrays(GL_QUADS, 0, 4);
        glDisableClientState(GL_COLOR_ARRAY);
        glColor4fv(baseColor);
    }

    void J2D::FillRoundedRect(const Color4& color, const Vector2& pos, const Vector2& size, float radius, unsigned int subdivisions) {
        if (!inJ2D)
            Logger::Error("Drawing J2D element before J2D begin.");

        J2D::FillRect(color, {pos.x + radius, pos.y}, {size.x - 2 * radius, size.y});
        J2D::FillRect(color, {pos.x, pos.y + radius}, {size.x, size.y - 2 * radius});

        J2D::FillCircle(color, {pos.x + radius, pos.y + radius}, radius, subdivisions);
        J2D::FillCircle(color, {pos.x + size.x - radius, pos.y + radius}, radius, subdivisions);
        J2D::FillCircle(color, {pos.x + radius, pos.y + size.y - radius}, radius, subdivisions);
        J2D::FillCircle(color, {pos.x + size.x - radius, pos.y + size.y - radius}, radius, subdivisions);
    }

    void
    J2D::DrawRenderTargetAsSprite(const JGL::RenderTarget& rt, const Vector2& position, float rad_rotation, const Vector2& origin,
                                  const Vector2& scale, const Color4& color, Direction inversion) {

        //Correct for the render-target being upside-down.
        Direction d{};
        if (inversion == Direction::None)
            d = Direction::Vertical;
        else if (inversion == Direction::Horizontal)
            d = Direction::Horizontal | Direction::Vertical;
        else if (inversion& Direction::Horizontal && inversion& Direction::Vertical)
            d = Direction::Horizontal;

        //Change the blending mode such that the alpha doesn't get multiplied again.
        glBlendFunc(GL_ONE, GL_ONE_MINUS_SRC_ALPHA);
        //J2D::DrawSprite(*rt.GetJGLTexture(), position, rad_rotation, origin, scale, color, d);
        J2D::DrawPartialSprite(*rt.GetJGLTexture(), position, {0, 0}, rt.GetDimensions(), rad_rotation, origin, scale, color, d);
        glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
    }
    void J2D::DrawSprite(const Texture& texture, const Vector2& pos, float rad_rotation, const Vector2& origin,
                         const Vector2& scale, const Color4& color, Direction inversion) {
        if (!inJ2D)
            Logger::Error("Drawing J2D element before J2D begin.");


        const Vector2 size = texture.GetDimensions();

        std::array<Vector2, 4> textureCoordinates = {Vector2(0, 0), Vector2(0, 1), Vector2(1, 1), Vector2(1, 0)};

        // TODO: Kind of a mess, refactor to be more sensible later.
        // Factors in scaling and origin correctly.
        // i.e. to render at 2x size, from the center, at coords XY, use {2, 2} scale, and {0.5, 0.5} offset.
        const Vector2 offset = origin * size;
        Vector2 pos2 = pos;
        Vector2 scaled_size = scale * size;
        Vector2 size2 = scaled_size;
        float cos_theta = std::cos(rad_rotation);
        float sin_theta = std::sin(rad_rotation);

        std::array<Vector2, 4> vertices =
                {
                        pos2,                                 // Top-left vertex
                        {pos2.x, pos2.y + size2.y},            // Bottom-left
                        {pos2.x + size2.x, pos2.y + size2.y},   // Bottom-right
                        {pos2.x + size2.x, pos2.y}             // Top-right
                };

        //Rotate the vertices about the origin by float rad_rotation.
        if (rad_rotation != 0)
            for (auto& v: vertices)
                v = {(v.x - pos2.x - offset.x * scale.x) * cos_theta - (v.y - pos2.y - offset.y * scale.y) * sin_theta +
                     pos2.x + offset.x * scale.x,
                     (v.x - pos2.x - offset.x * scale.x) * sin_theta + (v.y - pos2.y - offset.y * scale.y) * cos_theta +
                     pos2.y + offset.y * scale.y
                    };


        if (inversion == Direction::Vertical)
            textureCoordinates = {Vector2(0, 1), Vector2(0, 0), Vector2(1, 0), Vector2(1, 1)};
        else if (inversion == Direction::Horizontal)
            textureCoordinates = {Vector2(1, 0), Vector2(1, 1), Vector2(0, 1), Vector2(0, 0)};
        else if ((inversion& Direction::Horizontal) && (inversion& Direction::Vertical))
            textureCoordinates = {Vector2(1, 1), Vector2(1, 0), Vector2(0, 0), Vector2(0, 1)};

        glColor4ubv(color.ptr());
        glBindTexture(GL_TEXTURE_2D, texture.GetGLTextureHandle());
        glVertexPointer(2, GL_FLOAT, sizeof(Vector2), vertices.data());
        glTexCoordPointer(2, GL_FLOAT, sizeof(Vector2), textureCoordinates.data());
        glDrawArrays(GL_QUADS, 0, 4);
        glBindTexture(GL_TEXTURE_2D, 0);
        glColor4fv(baseColor);
    }


    void J2D::DrawSprite(const Texture& texture, float positionX, float positionY, float rad_rotation, float originX,
                         float originY, float scaleX, float scaleY, const Color4& color, Direction inversion) {
        DrawSprite(texture,
                   {positionX, positionY},
                   rad_rotation,
                   {originX, originY},
                   {scaleX, scaleY},
                   color, inversion);
    }

    void J2D::DrawPartialSprite(const Texture& texture, const Vector2 &position, const Vector2& sub_texture_position,
                                const Vector2& sub_texture_size, float rad_rotation, const Vector2& origin,
                                const Vector2& scale, const Color4& color, Direction inversion) {
        if (!inJ2D)
            Logger::Error("Drawing J2D element before J2D begin.");

        const Vector2 textureSize = texture.GetDimensions();

        // Calculate texture coordinates (relative to the whole texture)
        std::array<GLfloat, 8> textureCoordinates = {
                sub_texture_position.x / textureSize.x,
                sub_texture_position.y / textureSize.y,
                sub_texture_position.x / textureSize.x,
                (sub_texture_position.y + sub_texture_size.y) / textureSize.y,
                (sub_texture_position.x + sub_texture_size.x) / textureSize.x,
                (sub_texture_position.y + sub_texture_size.y) / textureSize.y,
                (sub_texture_position.x + sub_texture_size.x) / textureSize.x,
                sub_texture_position.y / textureSize.y
        };

        if (inversion& Direction::Vertical)
            std::swap(textureCoordinates[1], textureCoordinates[3]),
            std::swap(textureCoordinates[5], textureCoordinates[7]);

        if (inversion& Direction::Horizontal)
            std::swap(textureCoordinates[0], textureCoordinates[6]),
            std::swap(textureCoordinates[2], textureCoordinates[4]);

        const Vector2 offset = origin * sub_texture_size;
        Vector2 pos2 = position;
        Vector2 scaled_size = scale * sub_texture_size;
        Vector2 size2 = scaled_size;
        float cos_theta = std::cos(rad_rotation);
        float sin_theta = std::sin(rad_rotation);

        std::array<Vector2, 4> vertices =
                {
                        pos2,                                // Top-left
                        {pos2.x, pos2.y + size2.y},          // Bottom-left
                        {pos2.x + size2.x, pos2.y + size2.y},// Bottom-right
                        {pos2.x + size2.x, pos2.y}           // Top-right
                };

        //Rotate the vertices about the origin by float rad_rotation.
        if (rad_rotation != 0)
            for (auto& v: vertices)
                v = {(v.x - pos2.x - offset.x * scale.x) * cos_theta - (v.y - pos2.y - offset.y * scale.y) * sin_theta +
                     pos2.x + offset.x * scale.x,
                     (v.x - pos2.x - offset.x * scale.x) * sin_theta + (v.y - pos2.y - offset.y * scale.y) * cos_theta +
                     pos2.y + offset.y * scale.y};

        glColor4ubv(color.ptr());
        glBindTexture(GL_TEXTURE_2D, texture.GetGLTextureHandle());
        glVertexPointer(2, GL_FLOAT, sizeof(Vector2), vertices.data());
        glTexCoordPointer(2, GL_FLOAT, 0, textureCoordinates.data());
        glDrawArrays(GL_QUADS, 0, 4);
        glBindTexture(GL_TEXTURE_2D, 0);
        glColor4fv(baseColor);
    }

    void
    J2D::DrawPartialSprite(const JGL::Texture& texture, float positionX, float positionY, float sub_texture_positionX,
                           float sub_texture_positionY, unsigned int sub_texture_sizeX,
                           unsigned int sub_texture_sizeY, float originX, float originY, float rad_rotation,
                           float scaleX, float scaleY, const Color4& color, JGL::Direction inversion) {

        J2D::DrawPartialSprite(texture, {positionX, positionY}, {sub_texture_positionX, sub_texture_positionY},
                               {(float) sub_texture_sizeX, (float) sub_texture_sizeY}, rad_rotation, {originX, originY},
                               {scaleX, scaleY}, color, inversion);
    }

    void J2D::DrawMirrorSprite(const Texture& texture, const Vector2& position, Direction mirror_axis, float rad_rotation, const Vector2& origin, const Vector2& scale, const Color4& color) {
        if (!inJ2D)
            Logger::Error("Drawing J2D element before J2D begin.");

        if (mirror_axis == Direction::None)
            Logger::Warning("Drawing non-mirrored sprite with J2D::DrawMirrorSprite?");

        glBindTexture(GL_TEXTURE_2D, texture.GetGLTextureHandle());
        Vector2 size = texture.GetDimensions();
        std::array<Vector2, 4> textureCoordinates  = {Vector2(0, 0), Vector2(0, 1), Vector2(1, 1), Vector2(1, 0)};

        if (mirror_axis == Direction::Horizontal)
            size.x *= 2,
            glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_MIRRORED_REPEAT),
            textureCoordinates = {Vector2(0, 0), Vector2(0, 1), Vector2(2, 1), Vector2(2, 0)};

        else if (mirror_axis == Direction::Vertical)
            size.y *= 2,
            glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_MIRRORED_REPEAT),
            textureCoordinates = {Vector2(0, 0), Vector2(0, 2), Vector2(1, 2), Vector2(1, 0)};

        else if ((mirror_axis& Direction::Horizontal) && (mirror_axis& Direction::Vertical))
            size *= 2,
            textureCoordinates = {Vector2(0, 0), Vector2(0, 2), Vector2(2, 2), Vector2(2, 0)};

        const Vector2 offset = origin * size;
        Vector2 pos2 = position;
        Vector2 size2 = scale * size;
        float cos_theta = std::cos(rad_rotation);
        float sin_theta = std::sin(rad_rotation);

        std::array<Vector2, 4> vertices =
                {
                    pos2,
                    {pos2.x, pos2.y + size2.y},
                    {pos2.x + size2.x, pos2.y + size2.y},
                    {pos2.x + size2.x, pos2.y}
                };

        if (rad_rotation != 0)
            for (auto& v : vertices)
                v = {
                        (v.x - pos2.x - offset.x * scale.x) * cos_theta - (v.y - pos2.y - offset.y * scale.y) * sin_theta + pos2.x + offset.x * scale.x,
                        (v.x - pos2.x - offset.x * scale.x) * sin_theta + (v.y - pos2.y - offset.y * scale.y) * cos_theta + pos2.y + offset.y * scale.y
                };

        glColor4ubv(color.ptr());
        glVertexPointer(2, GL_FLOAT, sizeof(Vector2), vertices.data());
        glTexCoordPointer(2, GL_FLOAT, sizeof(Vector2), textureCoordinates.data());
        glDrawArrays(GL_QUADS, 0, 4);

        //Reset the wrapping mode.
        if (texture.GetWrappingMode() == TextureWrappingMode::CLAMP_TO_EDGE)
            glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_EDGE),
            glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_EDGE);
        else if (texture.GetWrappingMode() == TextureWrappingMode::REPEAT)
            glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_REPEAT),
            glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_REPEAT);
        else if (texture.GetWrappingMode() == TextureWrappingMode::CLAMP_TO_BORDER)
            glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_BORDER),
            glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_BORDER);

        glBindTexture(GL_TEXTURE_2D, 0);
        glColor4fv(baseColor);
    }

    void J2D::OutlineCircle(const Color4& color, const Vector2& center, float radius, unsigned int subdivisions, float thickness) {
        if (!inJ2D)
            Logger::Error("Drawing J2D element before J2D begin.");

        float step = (2.f * Math::Pi) / (float) subdivisions;
        std::vector<Vector2> vertices(subdivisions);
        GLfloat angle, x, y;

        int i = 0;
        for (angle = 0.0f; angle < (2.f * Math::Pi); angle += step) {
            x = radius * std::sin(angle) + center.x;
            y = radius * std::cos(angle) + center.y;
            if (i <= subdivisions)
                vertices[i] = {x, y};
            else
                vertices.emplace_back(x, y);
            i++;
        }

        glLineWidth(thickness);
        glColor4ubv(color.ptr());
        glVertexPointer(2, GL_FLOAT, sizeof(Vector2), vertices.data());
        glDrawArrays(GL_LINE_LOOP, 0, (int) vertices.size());
        glColor4fv(baseColor);
    }

    void J2D::FillCircle(const Color4& color, const Vector2& center, float radius, unsigned int subdivisions) {
        if (!inJ2D)
            Logger::Error("Drawing J2D element before J2D begin.");

        GLfloat angle, x, y;
        float step =  (2.f * Math::Pi) / (float) subdivisions;
        std::vector<Vector2> vertices(subdivisions);

        /* Most of the time the vector "vertices" is either the same size or size + 1 of the number of subdivisions.
         * Because a float is a decimal, It'd take too long to get rid of the emplace-back making us
         * wait around for the container to resize. This gets rid of it for what we can guarantee. */
        int i = 0;
        for (angle = 0.0f; angle < (2.f * Math::Pi); angle += step) {
            x = radius * std::sin(angle) + center.x;
            y = radius * std::cos(angle) + center.y;
            if (i <= subdivisions)
                vertices[i] = {x, y};
            else
                vertices.emplace_back(x, y);
            i++;
        }

        glColor4ubv(color.ptr());
        glVertexPointer(2, GL_FLOAT, sizeof(Vector2), vertices.data());
        glDrawArrays(GL_TRIANGLE_FAN, 0, (int) vertices.size());
        glColor4fv(baseColor);
    }

    void J2D::OutlineTriangle(const Color4& color, const Triangle2D& tri, float thickness) {
        if (!inJ2D)
            Logger::Error("Drawing J2D element before J2D begin.");

        Vector2 vertices[] = {{tri.A.x, tri.A.y}, {tri.B.x, tri.B.y}, {tri.C.x, tri.C.y}};

        glLineWidth(thickness);
        glColor4ubv(color.ptr());
        glVertexPointer(2, GL_FLOAT, sizeof(Vector2), vertices);
        glDrawArrays(GL_LINE_LOOP, 0, 3);
        glColor4fv(baseColor);
    }

    void J2D::FillTriangle(const Color4& color, const Triangle2D& tri) {
        if (!inJ2D)
            Logger::Error("Drawing J2D element before J2D begin.");

        Vector2 vertices[] = {{tri.A.x, tri.A.y}, {tri.B.x, tri.B.y}, {tri.C.x, tri.C.y}};

        glColor4ubv(color.ptr());
        glVertexPointer(2, GL_FLOAT, sizeof(Vector2), vertices);
        glDrawArrays(GL_TRIANGLES, 0, 3);
        glColor4fv(baseColor);
    }

    void J2D::FillGradientTriangle(const Color4& a_color, const Color4& b_color, const Color4& c_color, const Triangle2D& tri) {
        if (!inJ2D)
            Logger::Error("Drawing J2D element before J2D begin.");

        Vector2 vertices[] = {{tri.A.x, tri.A.y}, {tri.B.x, tri.B.y}, {tri.C.x, tri.C.y}};
        GLfloat colors[] = {a_color.RedChannelNormalized(), a_color.GreenChannelNormalized(), a_color.BlueChannelNormalized(), a_color.AlphaChannelNormalized(),
                            b_color.RedChannelNormalized(),b_color.GreenChannelNormalized(), b_color.BlueChannelNormalized(), b_color.AlphaChannelNormalized(),
                            c_color.RedChannelNormalized(), c_color.GreenChannelNormalized(), c_color.BlueChannelNormalized(),c_color.AlphaChannelNormalized()};

        glEnableClientState(GL_COLOR_ARRAY);
        glColorPointer(4, GL_FLOAT, sizeof(Color4), colors);
        glVertexPointer(2, GL_FLOAT, sizeof(Vector2), vertices);
        glDrawArrays(GL_TRIANGLES, 0, 3);
        glDisableClientState(GL_COLOR_ARRAY);
        glColor4fv(baseColor);
    }

    //TODO render all in once pass with GL_LINE_LOOP instead of separate lines.
    void J2D::DrawCubicBezierCurve(const Color4 &color, const Vector2& controlA, const Vector2& pointA, const Vector2& pointB, const Vector2& controlB,
                                   int subdivisions, float thickness) {


        Vector2 last = controlA;
        const Vector2& first = controlB;
        for (int i = 0; i < subdivisions; ++i)
        {
            float alpha = (float) i / (float) subdivisions;
            Vector2 step = J3ML::Algorithm::Bezier(alpha, controlA, pointA, pointB, controlB);
            DrawLine(color, last, step, thickness);
            last = step;
        }

        // Have to manually draw the last segment of the curve.
        DrawLine(color, last, first, thickness);
    }

    void J2D::OutlinePolygon(const Color4 &color, const std::vector<Vector2>& points, float thickness) {
        if (!inJ2D)
            Logger::Error("Drawing J2D element before J2D begin.");

        if (points.front() != points.back())
            throw std::runtime_error("J2D::OutlinePolygon: The first point and the last point must connect.");

        glLineWidth(thickness);
        glColor4ubv(color.ptr());
        glVertexPointer(2, GL_FLOAT, sizeof(Vector2), points.data());
        glDrawArrays(GL_LINE_LOOP, 0, (int) points.size());
        glColor4fv(baseColor);
    }

#pragma endregion

#pragma region J3D
    std::array<GLfloat, 16> OpenGLPerspectiveProjectionRH(float fovY, float aspect, float z_near, float z_far) {
        std::array<GLfloat, 16> result{};
        GLfloat f = 1.0f / std::tan(fovY * 0.5f * Math::Pi / 180.0f);
        result[0] = f / aspect;
        result[5] = f;
        result[10] = (z_far + z_near) / (z_near - z_far);
        result[11] = -1.0f;
        result[14] = (2.0f * z_far * z_near) / (z_near - z_far);
        return result;
    }

    bool j3d_initialized = false;
    float j3d_far_plane = 0;
    float j3d_fov = 0;

    void J3D::Init(const J3ML::LinearAlgebra::Vector2& window_size, float fov, float far_plane) {
        wS = window_size;
        j3d_far_plane = far_plane;
        j3d_fov = fov;
        j3d_initialized = true;
    }

    void J3D::ChangeFOV(float fov) {
        j3d_fov = fov;
    }

    void J3D::ChangeFarPlane(float far_plane) {
        j3d_far_plane = far_plane;
    }

    void J3D::Begin() {
        if (!j3d_initialized)
            throw std::runtime_error("You have to run J3D::Init before rendering 3D elements.");

        auto aspect = (float) wS.x / (float) wS.y;
        glMatrixMode(GL_PROJECTION);
        glLoadIdentity();
        glMultMatrixf(OpenGLPerspectiveProjectionRH(j3d_fov, aspect, 0.001, j3d_far_plane).data());
        glMatrixMode(GL_MODELVIEW);

        //Get what the draw color was before we did anything.
        glGetFloatv(GL_CURRENT_COLOR, oldColor);
        glColor4fv(baseColor);

        wasDepthTestEnabled = false;
        if (glIsEnabled(GL_DEPTH_TEST))
            wasDepthTestEnabled = true,
            glDisable(GL_DEPTH_TEST);

        wasVertexArraysEnabled = false;
        if (!glIsEnabled(GL_VERTEX_ARRAY))
            wasVertexArraysEnabled = false,
            glEnableClientState(GL_VERTEX_ARRAY);

        if (!glIsEnabled(GL_TEXTURE_COORD_ARRAY))
            wasTextureCoordArrayEnabled = false,
            glEnableClientState(GL_TEXTURE_COORD_ARRAY);
        else
            wasTextureCoordArrayEnabled = true;


        wasTexture2DEnabled = true;
        if (!glIsEnabled(GL_TEXTURE_2D))
            wasTexture2DEnabled = false,
            glEnable(GL_TEXTURE_2D);

        wasCullFaceEnabled = false;
        if (glIsEnabled(GL_CULL_FACE))
            wasCullFaceEnabled = true,
            glDisable(GL_CULL_FACE);

        if (!glIsEnabled(GL_BLEND))
            wasBlendEnabled = false,
            glEnable(GL_BLEND),
            glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
        else
            wasBlendEnabled = true;

        if (!inJ2D)
            inJ3D = true;
        else
            Logger::Error("Can't begin J3D context inside J2D context.");
    }

    void J3D::End() {
        if (wasDepthTestEnabled)
            glEnable(GL_DEPTH_TEST);

        if (!wasVertexArraysEnabled)
            glDisableClientState(GL_VERTEX_ARRAY);

        if (wasTexture2DEnabled)
            glDisable(GL_TEXTURE_2D);

        if (!wasBlendEnabled)
            glDisable(GL_BLEND);

        if (wasCullFaceEnabled)
            glEnable(GL_CULL_FACE);

        if (!wasTextureCoordArrayEnabled)
            glDisableClientState(GL_TEXTURE_COORD_ARRAY);

        //Put the draw color back how it was before.
        glColor4fv(oldColor);
        inJ3D = false;
    }

    void J3D::DrawLine(const Color4& color, const Vector3& A, const Vector3& B, float thickness) {
        if (!inJ3D)
            Logger::Error("Attempt to Render J3D element before J3D begin.");

        Vector3 vertices[] = {A, B};

        glLineWidth(thickness);
        glColor4ubv(color.ptr());
        glVertexPointer(3, GL_FLOAT, sizeof(Vector3), vertices);
        glDrawArrays(GL_LINES, 0, 2);
        glColor4fv(baseColor);
    }
#pragma endregion
}
