#include <JGL/types/RenderTarget.h>
#include <jlog/Logger.hpp>
#include <stdexcept>

JGL::Texture* JGL::RenderTarget::GetJGLTexture() const {
    return texture;
}

GLuint JGL::RenderTarget::GetGLTextureHandle() const {
    return texture->GetGLTextureHandle();
}

GLuint JGL::RenderTarget::GetGLFramebufferObjectHandle() const {
    return framebuffer_object;
}

GLuint JGL::RenderTarget::GetGLDepthBufferHandle() const {
    return depth_buffer;
}

GLuint JGL::RenderTarget::GetActiveGLFramebufferHandle() {
    GLuint fbo;
    glGetIntegerv(GL_FRAMEBUFFER_BINDING, (GLint*) &fbo);
    return fbo;
}

void JGL::RenderTarget::SetActiveGLRenderTarget(const RenderTarget& render_target) {
    RenderTarget rt = render_target;
    glBindFramebuffer(GL_FRAMEBUFFER, rt.GetGLFramebufferObjectHandle());
    glViewport(0,0, rt.GetDimensions().x, rt.GetDimensions().y);
}

Vector2 JGL::RenderTarget::GetDimensions() const {
    return size;
}

void JGL::RenderTarget::Erase() {
    if (GetActiveGLFramebufferHandle() == framebuffer_object)
        jlog::Warning("Deleting the framebuffer that's currently in use?");

    texture->Erase();

    if (using_depth)
        glDeleteRenderbuffers(1, &depth_buffer);

    glDeleteFramebuffers(1, &framebuffer_object);
}

Color4 JGL::RenderTarget::GetClearColor() const {
    return clear_color;
}

JGL::RenderTarget::RenderTarget(const Vector2& size, const Color4& clear_color, bool use_depth) {
    GLuint current_fbo = GetActiveGLFramebufferHandle();
    GLint viewport[4] = {0, 0, 0, 0};
    glGetIntegerv(GL_VIEWPORT, viewport);
    //Textures behave strangely if they're not square aaaaaaaaaaaaa.
    unsigned int biggest;
    if (size.x >= size.y)
        biggest = size.x;
    else biggest = size.y;

    texture = new Texture(Vector2(biggest, biggest));
    glGenFramebuffers(1, &framebuffer_object);
    glBindFramebuffer(GL_FRAMEBUFFER, framebuffer_object);
    glViewport(0,0, size.x, size.y);
    glFramebufferTexture2D(GL_FRAMEBUFFER, GL_COLOR_ATTACHMENT0, GL_TEXTURE_2D, texture->GetGLTextureHandle(), 0);

    if (use_depth) {
        GLuint depthBuffer;
        glGenRenderbuffers(1, &depthBuffer);
        glBindRenderbuffer(GL_RENDERBUFFER, depthBuffer);
        glRenderbufferStorage(GL_RENDERBUFFER, GL_DEPTH_COMPONENT, biggest, biggest);
        glFramebufferRenderbuffer(GL_FRAMEBUFFER, GL_DEPTH_ATTACHMENT, GL_RENDERBUFFER, depthBuffer);
        glClear(GL_DEPTH_BUFFER_BIT);
        using_depth = true;
    }

    GLfloat old_clear_color[4];
    glGetFloatv(GL_COLOR_CLEAR_VALUE, old_clear_color);
    glClearColor(clear_color.RedChannelNormalized(), clear_color.GreenChannelNormalized(), clear_color.BlueChannelNormalized(), clear_color.AlphaChannelNormalized());
    glClear(GL_COLOR_BUFFER_BIT);
    glClearColor(old_clear_color[0], old_clear_color[1], old_clear_color[2], old_clear_color[3]);

    if (glCheckFramebufferStatus(GL_FRAMEBUFFER) != GL_FRAMEBUFFER_COMPLETE)
        throw std::runtime_error("A new framebuffer could not be allocated.");

    glBindFramebuffer(GL_FRAMEBUFFER, current_fbo);
    glViewport(viewport[0], viewport[1], viewport[2], viewport[3]);
    this->clear_color = clear_color;
    this->size = size;
}

std::vector<GLfloat> JGL::RenderTarget::GetData() const {
    std::vector<GLfloat> data(GetDimensions().x * GetDimensions().y * 4);
    GLuint current_fbo = GetActiveGLFramebufferHandle();

    glBindFramebuffer(GL_FRAMEBUFFER, framebuffer_object);
    glReadPixels(0, 0, GetDimensions().x, GetDimensions().y, GL_RGBA, GL_FLOAT, data.data());
    glBindFramebuffer(GL_FRAMEBUFFER, current_fbo);
    return data;

}
